#include <imsls.h>
#include <imsls.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

typedef struct
{
   int n_queens;
} inputArgs;
int main(){
   int i, j, k;        /* index variables             */
   int n = 500;        /* population size             */
   int n_generations;  /* number of generations       */
   int n_queens = 25;  /* number of nominal phenotypes*/
   int n_categories[25];
   float maxFit;       /* maximum fitness hurdle      */
   float* genStats;    /* generation statistics       */
   static float queensFitness(
      Imsls_f_individual* individual, inputArgs* input);
   inputArgs* parameters;
   Imsls_f_chromosome* chromosome;     /* chromosome */
   Imsls_f_individual* best_individual;/* optimum    */
   Imsls_f_population* population;     /* population */
   /**************************************************/
   /* In this example the user function is thread    */
   /* safe.  Let CNL know it is safe, which allows   *
   /* genetic algorithm to run in parallel, if that  */
   /* capability exists on the user computer.        */
   imsls_omp_options(IMSLS_SET_FUNCTIONS_THREAD_SAFE, 1, 0);
   imsls_random_seed_set(12345);
   maxFit = n_queens - 0.5;
   for(i=0; i<n_queens; i++) n_categories[i] = n_queens;
   chromosome      = imsls_f_ga_chromosome(
            IMSLS_NOMINAL, n_queens, n_categories, 0);
   
   parameters = (inputArgs*) malloc(sizeof(inputArgs));
   parameters->n_queens = n_queens;
   population = imsls_f_ga_random_population(n, 
            chromosome,
            IMSLS_PMX_CROSSOVER,
            IMSLS_FITNESS_FCN_WITH_PARMS, 
            queensFitness, parameters, 0);
   best_individual = imsls_f_genetic_algorithm(
         NULL, population, 
         IMSLS_FITNESS_FCN_WITH_PARMS, queensFitness, parameters,
         IMSLS_PRINT_LEVEL, IMSLS_FINAL,
         IMSLS_PMX_CROSSOVER,
         IMSLS_LINEAR_SCALING, 2.0,
         IMSLS_CROSSOVER_PROB, 0.7,
         IMSLS_MUTATION_PROB, 0.01,
         IMSLS_MAX_GENERATIONS, 10000,
         IMSLS_MAX_FITNESS, maxFit,
         IMSLS_GENERATION_STATS, &genStats,
         IMSLS_N_GENERATIONS, &n_generations, 0);

   printf("GENERATION STATISTICS\n");
   printf("Total Number of Generations: %d\n\n", n_generations);
   printf("Generation  Max. Fit. Min. Fit.  Avg. Fit.    CV\n");
   printf("*************************************************\n");
   for(i=0; i<=n_generations; i+=25){
      printf("Gen. %3d: %8.0f %8.0f %12.2f %8.2f\n",
         i, genStats[4*i], genStats[4*i+1], genStats[4*i+2],
         100*genStats[4*i+3]/genStats[4*i+2]);
   } 
   i = n_generations;
   printf("Gen. %3d: %8.0f %8.0f %12.2f %8.2f\n",
         i, genStats[4*i], genStats[4*i+1], genStats[4*i+2],
         100*genStats[4*i+3]/genStats[4*i+2]);
   printf("*************************************************\n");
   printf("OPTIMUM SOLUTION:\n\nFitness:%4.0f\n\nChromosome:        ", 
      queensFitness(best_individual, parameters));
   for(i=0; i<best_individual->chromosome->c_length; i++) 
      printf("%3d", best_individual->chromosome->allele[i]);
   if(n_queens<100){
      printf("\n\nBoard Positions:  \n\n");
      for(i=0; i<n_queens; i++) printf("--");
      printf("-\n");
      for(i=0; i<n_queens; i++) {
         for(j=0; j<n_queens; j++){
            k = best_individual->nominalPhenotype[j];
            if(i==k) printf("|Q");
            else     printf("| ");
         }
         printf("|\n");
         for(k=0; k<n_queens; k++) printf("--");
         printf("-\n");
      }
      printf("\n\n");
   }
}
static float queensFitness(Imsls_f_individual* individual, inputArgs* input)
{
   int i=0, j=0, k=0; /* Index variables */
   int n_queens;
   float f = 0.0;     /* Fitness value   */
   n_queens = input->n_queens;
   f = 0;
   for(i=0; i<n_queens-1; i++){
      for(j=i+1; j<n_queens; j++){
         k = individual->chromosome->allele[i] -
            individual->chromosome->allele[j] ;
         k = abs(k);
         if (abs(i-j) == k) f++;
      }
   }
   f = n_queens - f;
   return f;
}
